from __future__ import print_function
from __future__ import unicode_literals
from dali.driver.base import DALIDriver
from dali.driver.base import SyncDALIDriver
from dali.frame import BackwardFrame
from dali.frame import ForwardFrame
from dali.gear.general import Off, DAPC, Up, Down, StepUp, StepDown, RecallMaxLevel, RecallMinLevel, StepDownAndOff, OnAndStepUp, \
    EnableDAPCSequence, GoToLastActiveLevel, GoToScene, Reset, StoreActualLevelInDTR0, \
    SetMaxLevel, SetMinLevel, SetSystemFailureLevel, SetPowerOnLevel,\
    SetFadeTime, SetExtendedFadeTime, RemoveFromScene, RemoveFromGroup,\
    QueryControlGearPresent, QueryLampFailure,\
    QueryLampPowerOn, QueryLimitError, QueryPhysicalMinimum, QueryPowerFailure,\
    QueryContentDTR1, QueryContentDTR2, QueryOperatingMode, QueryLightSourceType,\
    QueryMaxLevel, QueryPowerOnLevel, QuerySystemFailureLevel,\
    QueryFadeTimeFadeRate, QueryManufacturerSpecificMode, QueryNextDeviceType,\
    QueryExtendedFadeTime, QueryControlGearFailure, QuerySceneLevel,\
    QueryGroupsZeroToSeven, QueryGroupsEightToFifteen, Terminate,\
    ProgramShortAddress, VerifyShortAddress, QueryShortAddress, EnableDeviceType,\
    WriteMemoryLocation, WriteMemoryLocationNoReply, \
    DTR0, DTR1, DTR2, IdentifyDevice, SetFadeRate, SetScene, QueryContentDTR0, \
    QueryMinLevel, QueryRandomAddressH, QueryRandomAddressM, QueryRandomAddressL, \
    ReadMemoryLocation, QueryExtendedVersionNumber, Initialise, Randomise, Compare, Ping,\
    QueryStatus, QueryDeviceType, QueryActualLevel, AddToGroup, SearchaddrH, SearchaddrM, SearchaddrL, \
    SavePersistentVariables, SetOperatingMode, ResetMemoryBank, SetShortAddress, EnableWriteMemory,\
    QueryResetState, QueryMissingShortAddress, QueryVersionNumber
import logging
from remotearm import RemoteArm
from dali.address import Short, Broadcast, Group, BroadcastUnaddressed
from dali.bus import Bus
from time import sleep
import threading

DA_OPT_TWICE = 0x8
DA_OPT_17BIT = 0x4

# debug logging related
DRIVER_CONSTRUCT = 0x0
DRIVER_EXTRACT = 0x1
_exco_str = {
    DRIVER_CONSTRUCT: 'CONSTRUCT',
    DRIVER_EXTRACT: 'EXTRACT',
}

COMMAND_LIST = {
    1: (DAPC, True),
    2: (Off, False),
    3: (Up, False),
    4: (Down, False),
    5: (StepUp, False),
    6: (StepDown, False),
    7: (RecallMaxLevel, False),
    8: (RecallMinLevel, False),
    9: (StepDownAndOff, False),
    10: (OnAndStepUp, False),
    11: (EnableDAPCSequence, False),
    12: (GoToLastActiveLevel, False),
    13: (GoToScene, True),
    14: (Reset, False),
    15: (StoreActualLevelInDTR0, False),
    16: (SavePersistentVariables, False),
    17: (SetOperatingMode, False),
    18: (ResetMemoryBank, False),
    19: (IdentifyDevice, False),
    20: (SetMaxLevel, False),
    21: (SetMinLevel, False),
    22: (SetSystemFailureLevel, False),
    23: (SetPowerOnLevel, False),
    24: (SetFadeTime, False),
    25: (SetFadeRate, False),
    26: (SetExtendedFadeTime, False),
    27: (SetScene, True),
    28: (RemoveFromScene, True),
    29: (AddToGroup, True),
    30: (RemoveFromGroup, True),
    31: (SetShortAddress, False),
    32: (EnableWriteMemory, False),
    33: (QueryStatus, False),
    34: (QueryControlGearPresent, False),
    35: (QueryLampFailure, False),
    36: (QueryLampPowerOn, False),
    37: (QueryLimitError, False),
    38: (QueryResetState, False),
    39: (QueryMissingShortAddress, False),
    40: (QueryVersionNumber, False),
    41: (QueryContentDTR0, False),
    42: (QueryDeviceType, False),
    43: (QueryPhysicalMinimum, False),
    44: (QueryPowerFailure, False),
    45: (QueryContentDTR1, False),
    46: (QueryContentDTR2, False),
    47: (QueryOperatingMode, False),
    48: (QueryLightSourceType, False),
    49: (QueryActualLevel, False),
    50: (QueryMaxLevel, False),
    51: (QueryMinLevel, False),
    52: (QueryPowerOnLevel, False),
    53: (QuerySystemFailureLevel, False),
    54: (QueryFadeTimeFadeRate, False),
    55: (QueryManufacturerSpecificMode, False),
    56: (QueryNextDeviceType, False),
    57: (QueryExtendedFadeTime, False),
    58: (QueryControlGearFailure, False),
    59: (QuerySceneLevel, True),
    60: (QueryGroupsZeroToSeven, False),
    61: (QueryGroupsEightToFifteen, False),
    62: (QueryRandomAddressH, False),
    63: (QueryRandomAddressM, False),
    64: (QueryRandomAddressL, False),
    65: (ReadMemoryLocation, False),
    66: (QueryExtendedVersionNumber, False),
    100: (Terminate, False),
    101: (DTR0, True),
    102: (Initialise, False),
    103: (Randomise, False),
    104: (Compare, False),
    105: (Ping, False),
    106: (SearchaddrH, True),
    107: (SearchaddrM, True),
    108: (SearchaddrL, True),
    109: (ProgramShortAddress, False),
    110: (VerifyShortAddress, False),
    111: (QueryShortAddress, False),
    112: (EnableDeviceType, True),
    113: (DTR1, True),
    114: (DTR2, True),
    115: (WriteMemoryLocation, True),
    116: (WriteMemoryLocationNoReply, True)
}


def _log_frame(logger, exco, opt, ad, cm, st):
    msg = (
        '{}\n'
        '    Type: {}\n'
        '    Address: {}\n'
        '    Command: {}\n'
    )
    logger.info(msg.format(
        _exco_str[exco],
        hex(opt),
        hex(ad),
        hex(cm),
    ))

class DALINoResponse(object):

    def __repr__(self):
        return 'NO_RESPONSE'

    __str__ = __repr__

DALI_NO_RESPONSE = DALINoResponse()

class UnipiDALIDriver(DALIDriver):
    """``DALIDriver`` implementation for UniPi DALI ModBus device.
    """
    debug = False
    logger = logging.getLogger('UnipiDALIDriver')
    # next sequence number
    _next_sn = 1

    def construct(self, command):
        """Data expected by DALI Modbus 
             Reg_1 - 16bit - length, ...
             Reg_2 - 16bit - address,command
        """
        frame = command.frame
        if len(frame) == 16:
            opt = 0x2
            if command.is_config: opt |= DA_OPT_TWICE
            ad, cm1 = frame.as_byte_sequence
            reg2 = (ad << 8) | cm1
            reg1 = opt << 8
        elif len(frame) == 24:
            opt = 0x3
            if command.is_config: opt |= DA_OPT_TWICE
            ad, cm1, cm2 = frame.as_byte_sequence
            reg1 = (opt << 8) | ad
            reg2 = (cm1 << 8) | cm2
        else:
            raise ValueError('Unknown frame length: {}'.format(len(frame)))

        if self.debug:
            _log_frame(
                self.logger, DRIVER_CONSTRUCT, opt, ad, cm1,0)
        return (reg1, reg2)


    def extract(self, data):
        """ ----- """
        if data[0] == 0x100:
            return BackwardFrame(data[1])
        if data[0] == 0x200:
            return ForwardFrame(16, [data[1] >> 8, data[1] & 0xff])
        else:
            return DALI_NO_RESPONSE

    def _get_sn(self):
        """Get next sequence number."""
        sn = self._next_sn
        if sn > 255:
            sn = self._next_sn = 1
        else:
            self._next_sn += 1
        return sn


class SyncUnipiDALIDriver(UnipiDALIDriver, SyncDALIDriver):
    """Synchronous implementation for UnipiDali    """

    def __init__(self, bus=0, unit=1):
        self.backend = RemoteArm('127.0.0.1', unit=unit)
        self.bus = bus
        self._sendreg = 13 + 2 * bus
        self._recvreg = 1 + 3 * bus
        self._fereg = 38 + int(bus / 2)

    def send(self, command, timeout=None):
        sleep(0.05)
        registers = self.construct(command)
        print("TFC %s" % (str(command)))
        self.backend.write_regs(self._sendreg, registers)
        if command.is_config:
            sleep(0.05)
            self.backend.write_regs(self._sendreg, registers)
        counter1, = self.backend.read_regs(self._recvreg,1)
        fe_counter1, = self.backend.read_regs(self._fereg,1)
        try:
            needsResponse = command._response is not None
            frame = None
            if needsResponse: # wait max 6*11ms
                for i in range(6): 
                    sleep(0.011)
                    counter2, reg1, reg2 = self.backend.read_regs(self._recvreg, 3)
                    if (counter1 != counter2):
                        frame = self.extract((reg1, reg2))
                        if isinstance(frame, BackwardFrame):
                            if command.response:
                                return command.response(frame)
                    if hasattr(command, "_cmdval") and command._cmdval == 0xA9:
                        fe_counter2, = self.backend.read_regs(self._fereg, 1)
                        if fe_counter1 != fe_counter2:
                            return command.response(BackwardFrame(0xFF))
                return command.response(None)
        except BaseException, E:
            print(str(E))
        return DALI_NO_RESPONSE

class UniPiLightBus(object):
    driver = None
    internal_bus = None
    local_modbus_client = None
    tx_thread = None
    tx_address = 0
    tx_address_dict = {}
    tx_argument_dict = {}
    tx_argument = 0
    tx_func_queue = []
    opt_value_dict = {}
    
    def __init__(self, client, bus=0, unit=1):
        self.local_modbus_client = client
        self.driver = SyncUnipiDALIDriver(bus=bus, unit=unit)
        self.internal_bus = Bus('%d' % (bus+1), self.driver) 
        self.tx_thread = None
        self.tx_address = 0
        self.tx_address_dict = {}
        self.tx_argument_dict = {}
        self.tx_argument = 0
        self.tx_func_queue = []
        self.opt_value_dict = {}

    def scan(self):
        self.internal_bus.assign_short_addresses()
        
    def is_address_broadcast(self, qual_address):
        if qual_address >= 200 and qual_address < 203:
            return True
        return False
    
    def is_address_short(self, qual_address):
        if qual_address >= 0 and qual_address < 100:
            return True
        return False

    def is_address_group(self, qual_address):
        if qual_address >= 100 and qual_address < 200:
            return True
        return False
        
    def do_work(self, work):
        if len(self.tx_func_queue) < 128:
            self.tx_func_queue.append(work)
        self.wakeup_thread()
            
    def wakeup_thread(self):
        if not self.is_thread_active():
            self.tx_thread = threading.Thread(target=self.execute_queue, args=[])
            self.tx_thread.start()
        
    def execute_queue(self):
        while len(self.tx_func_queue) > 0:
            print("TXFQL: %d" % (len(self.tx_func_queue)))
            tx_func = self.tx_func_queue[0][0]
            tx_func_args = self.tx_func_queue[0][1]
            print(str(tx_func_args))
            if len(tx_func_args) == 0:
                tx_func()
            elif len(tx_func_args) == 1:
                tx_func(tx_func_args[0])
            elif len(tx_func_args) == 2:
                tx_func(tx_func_args[0], tx_func_args[1])
            elif len(tx_func_args) == 3:
                tx_func(tx_func_args[0], tx_func_args[1], tx_func_args[2])
            elif len(tx_func_args) == 4:
                tx_func(tx_func_args[0], tx_func_args[1], tx_func_args[2], tx_func_args[3])
            elif len(tx_func_args) == 5:
                tx_func(tx_func_args[0], tx_func_args[1], tx_func_args[2], tx_func_args[3], tx_func_args[4])
            self.tx_func_queue = self.tx_func_queue[1:]
        
    def construct_address(self, qual_address):
        if self.is_address_short(qual_address):
            return Short(qual_address - 100)
        elif self.is_address_group(qual_address):
            return Group(qual_address - 200)
        elif self.is_address_broadcast(qual_address):
            if qual_address == 1:
                return BroadcastUnaddressed()
            elif qual_address == 2:
                return Broadcast()
        else:
            return qual_address
        
    def send_command(self, command):
        self.driver.send(command)

    def is_thread_active(self):
        if not self.tx_thread or not self.tx_thread.isAlive():
            return False
        return True

    def perform_duplicate_write(self, register, value, address):
        sleep(0.011)
        if address in self.opt_value_dict.keys():
            self.local_modbus_client.write_register(register - 1, self.opt_value_dict[address])
        self.local_modbus_client.write_register(register, value)
        sleep(0.05)
        if address in self.opt_value_dict.keys():
            self.local_modbus_client.write_register(register - 1, self.opt_value_dict[address])
        self.local_modbus_client.write_register(register, value)
        
    def perform_single_write(self, register, value, address):
        sleep(0.011)
        if address in self.opt_value_dict.keys():
            self.local_modbus_client.write_register(register - 1, self.opt_value_dict[address])
        self.local_modbus_client.write_register(register, value)
        
    def perform_command(self, command, address, argument):
        sleep(0.005)
        if address == 300:
            if argument is not None:
                self.driver.send(command(argument))
            else:
                self.driver.send(command())
        else:     
            if argument is not None:
                self.driver.send(command(address, argument))
            else:
                self.driver.send(command(address))
        
    def get_command_name(self, cid):
        ret = None
        try:
            if cid in COMMAND_LIST.keys():
                ret = COMMAND_LIST[cid][0]
        except BaseException, E:
            print("Exc: " + str(E))
        return ret
    
    def get_tx_address(self, reg_address):
        if reg_address in self.tx_address_dict.keys():
            return self.tx_address_dict[reg_address]
        return 0
    
    def get_tx_argument(self, reg_address):
        if reg_address in self.tx_argument_dict.keys():
            return self.tx_argument_dict[reg_address]
        return 0
        
    def command_has_argument(self, cid):
        ret = False
        try:
            if cid in COMMAND_LIST.keys():
                ret = COMMAND_LIST[cid][1]
        except BaseException, E:
            print("Exc: " + str(E))
        return ret